from __future__ import annotations

import pytest

from dissect.evidence import exception


@pytest.mark.parametrize(
    ("exc", "std"),
    [
        (exception.FileNotFoundError, FileNotFoundError),
        (exception.IsADirectoryError, IsADirectoryError),
        (exception.NotADirectoryError, NotADirectoryError),
    ],
)
def test_filesystem_error_subclass(exc: exception.Error, std: Exception) -> None:
    assert issubclass(exc, std)
    assert isinstance(exc(), std)

    with pytest.raises(std):
        raise exc()



# Generated by cstruct-stubgen
from typing import BinaryIO, TypeAlias, overload

import dissect.cstruct as __cs__

class _c_ewf(__cs__.cstruct):
    class MediaType(__cs__.Enum):
        Removable = ...
        Fixed = ...
        Optical = ...
        Logical = ...
        RAM = ...

    class MediaFlags(__cs__.Enum):
        Image = ...
        Physical = ...
        Fastbloc = ...
        Tablaeu = ...

    class CompressionLevel(__cs__.Enum):
        # None = ...
        Good = ...
        Best = ...

    class EWFHeader(__cs__.Structure):
        signature: __cs__.CharArray
        fields_start: _c_ewf.uint8
        segment_number: _c_ewf.uint16
        fields_end: _c_ewf.uint16
        @overload
        def __init__(
            self,
            signature: __cs__.CharArray | None = ...,
            fields_start: _c_ewf.uint8 | None = ...,
            segment_number: _c_ewf.uint16 | None = ...,
            fields_end: _c_ewf.uint16 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class EWFSectionDescriptor(__cs__.Structure):
        type: __cs__.CharArray
        next: _c_ewf.uint64
        size: _c_ewf.uint64
        pad: __cs__.Array[_c_ewf.uint8]
        checksum: _c_ewf.uint32
        @overload
        def __init__(
            self,
            type: __cs__.CharArray | None = ...,
            next: _c_ewf.uint64 | None = ...,
            size: _c_ewf.uint64 | None = ...,
            pad: __cs__.Array[_c_ewf.uint8] | None = ...,
            checksum: _c_ewf.uint32 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class EWFVolumeSectionSpec(__cs__.Structure):
        reserved_1: _c_ewf.uint32
        chunk_count: _c_ewf.uint32
        sector_count: _c_ewf.uint32
        sector_size: _c_ewf.uint32
        total_sector_count: _c_ewf.uint32
        reserved: __cs__.Array[_c_ewf.uint8]
        pad: __cs__.Array[_c_ewf.uint8]
        signature: __cs__.CharArray
        checksum: _c_ewf.uint32
        @overload
        def __init__(
            self,
            reserved_1: _c_ewf.uint32 | None = ...,
            chunk_count: _c_ewf.uint32 | None = ...,
            sector_count: _c_ewf.uint32 | None = ...,
            sector_size: _c_ewf.uint32 | None = ...,
            total_sector_count: _c_ewf.uint32 | None = ...,
            reserved: __cs__.Array[_c_ewf.uint8] | None = ...,
            pad: __cs__.Array[_c_ewf.uint8] | None = ...,
            signature: __cs__.CharArray | None = ...,
            checksum: _c_ewf.uint32 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class EWFVolumeSection(__cs__.Structure):
        media_type: _c_ewf.MediaType
        reserved_1: __cs__.Array[_c_ewf.uint8]
        chunk_count: _c_ewf.uint32
        sector_count: _c_ewf.uint32
        sector_size: _c_ewf.uint32
        total_sector_count: _c_ewf.uint64
        num_cylinders: _c_ewf.uint32
        num_heads: _c_ewf.uint32
        num_sectors: _c_ewf.uint32
        media_flags: _c_ewf.uint8
        unknown_1: __cs__.Array[_c_ewf.uint8]
        palm_start_sector: _c_ewf.uint32
        unknown_2: _c_ewf.uint32
        smart_start_sector: _c_ewf.uint32
        compression_level: _c_ewf.CompressionLevel
        unknown_3: __cs__.Array[_c_ewf.uint8]
        error_granularity: _c_ewf.uint32
        unknown_4: _c_ewf.uint32
        uuid: __cs__.Array[_c_ewf.uint8]
        pad: __cs__.Array[_c_ewf.uint8]
        signature: __cs__.CharArray
        checksum: _c_ewf.uint32
        @overload
        def __init__(
            self,
            media_type: _c_ewf.MediaType | None = ...,
            reserved_1: __cs__.Array[_c_ewf.uint8] | None = ...,
            chunk_count: _c_ewf.uint32 | None = ...,
            sector_count: _c_ewf.uint32 | None = ...,
            sector_size: _c_ewf.uint32 | None = ...,
            total_sector_count: _c_ewf.uint64 | None = ...,
            num_cylinders: _c_ewf.uint32 | None = ...,
            num_heads: _c_ewf.uint32 | None = ...,
            num_sectors: _c_ewf.uint32 | None = ...,
            media_flags: _c_ewf.uint8 | None = ...,
            unknown_1: __cs__.Array[_c_ewf.uint8] | None = ...,
            palm_start_sector: _c_ewf.uint32 | None = ...,
            unknown_2: _c_ewf.uint32 | None = ...,
            smart_start_sector: _c_ewf.uint32 | None = ...,
            compression_level: _c_ewf.CompressionLevel | None = ...,
            unknown_3: __cs__.Array[_c_ewf.uint8] | None = ...,
            error_granularity: _c_ewf.uint32 | None = ...,
            unknown_4: _c_ewf.uint32 | None = ...,
            uuid: __cs__.Array[_c_ewf.uint8] | None = ...,
            pad: __cs__.Array[_c_ewf.uint8] | None = ...,
            signature: __cs__.CharArray | None = ...,
            checksum: _c_ewf.uint32 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class EWFTableSection(__cs__.Structure):
        num_entries: _c_ewf.uint32
        _: _c_ewf.uint32
        base_offset: _c_ewf.uint64
        checksum: _c_ewf.uint32
        entries: __cs__.Array[_c_ewf.uint32]
        @overload
        def __init__(
            self,
            num_entries: _c_ewf.uint32 | None = ...,
            _: _c_ewf.uint32 | None = ...,
            base_offset: _c_ewf.uint64 | None = ...,
            checksum: _c_ewf.uint32 | None = ...,
            entries: __cs__.Array[_c_ewf.uint32] | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

# Technically `c_ewf` is an instance of `_c_ewf`, but then we can't use it in type hints
c_ewf: TypeAlias = _c_ewf



from __future__ import annotations

from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING, BinaryIO

import pytest

from dissect.evidence.asdf import AsdfWriter
from tests._utils import absolute_path

if TYPE_CHECKING:
    from collections.abc import Iterator


def open_data(name: str) -> Iterator[BinaryIO]:
    with (Path(__file__).parent / name).open("rb") as fh:
        yield fh


@pytest.fixture
def ad1_basic() -> Iterator[BinaryIO]:
    yield from open_data("_data/ad1/test.ad1")


@pytest.fixture
def ad1_long() -> Iterator[BinaryIO]:
    yield from open_data("_data/ad1/long.ad1")


@pytest.fixture
def ad1_compressed() -> Iterator[BinaryIO]:
    yield from open_data("_data/ad1/compressed.ad1")


@pytest.fixture
def ad1_segmented() -> list[Path]:
    return [
        absolute_path("_data/ad1/pcbje/text-and-pictures.ad1"),
        absolute_path("_data/ad1/pcbje/text-and-pictures.ad2"),
        absolute_path("_data/ad1/pcbje/text-and-pictures.ad3"),
        absolute_path("_data/ad1/pcbje/text-and-pictures.ad4"),
    ]


@pytest.fixture
def ad1_encrypted_passphrase() -> list[Path]:
    return [
        absolute_path("_data/ad1/encrypted-passphrase/encrypted.ad1"),
        absolute_path("_data/ad1/encrypted-passphrase/encrypted.ad2"),
        absolute_path("_data/ad1/encrypted-passphrase/encrypted.ad3"),
        absolute_path("_data/ad1/encrypted-passphrase/encrypted.ad4"),
        absolute_path("_data/ad1/encrypted-passphrase/encrypted.ad5"),
        absolute_path("_data/ad1/encrypted-passphrase/encrypted.ad6"),
        absolute_path("_data/ad1/encrypted-passphrase/encrypted.ad7"),
        absolute_path("_data/ad1/encrypted-passphrase/encrypted.ad8"),
        absolute_path("_data/ad1/encrypted-passphrase/encrypted.ad9"),
        absolute_path("_data/ad1/encrypted-passphrase/encrypted.ad10"),
        absolute_path("_data/ad1/encrypted-passphrase/encrypted.ad11"),
        absolute_path("_data/ad1/encrypted-passphrase/encrypted.ad12"),
        absolute_path("_data/ad1/encrypted-passphrase/encrypted.ad13"),
    ]


@pytest.fixture
def ad1_encrypted_certificate() -> list[Path]:
    return [
        absolute_path("_data/ad1/encrypted-certificate/encrypted.ad1"),
        absolute_path("_data/ad1/encrypted-certificate/encrypted.ad2"),
        absolute_path("_data/ad1/encrypted-certificate/encrypted.ad3"),
        absolute_path("_data/ad1/encrypted-certificate/encrypted.ad4"),
        absolute_path("_data/ad1/encrypted-certificate/encrypted.ad5"),
        absolute_path("_data/ad1/encrypted-certificate/encrypted.ad6"),
        absolute_path("_data/ad1/encrypted-certificate/encrypted.ad7"),
    ]


@pytest.fixture
def ewf_data() -> Iterator[BinaryIO]:
    yield from open_data("_data/ewf/ewf.E01")


@pytest.fixture
def asdf_writer() -> AsdfWriter:
    def noop() -> None:
        pass

    fh = BytesIO()
    fh.close = noop  # Prevent clearing the buffer, we need it
    return AsdfWriter(fh)



from __future__ import annotations

import argparse
import logging
import shutil
from pathlib import Path

from dissect.evidence.ad1.ad1 import find_files as find_ad1_files
from dissect.evidence.adcrypt.adcrypt import ADCrypt, is_adcrypt
from dissect.evidence.ewf.ewf import find_files as find_ewf_files
from dissect.evidence.tools.util import catch_sigpipe

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]\t%(message)s")
log = logging.getLogger(__name__)


@catch_sigpipe
def main() -> None:
    help_formatter = argparse.ArgumentDefaultsHelpFormatter
    parser = argparse.ArgumentParser(
        prog="adcrypt",
        description="Decrypt E01 or AD1 ADCRYPT encrypted segment files.",
        fromfile_prefix_chars="@",
        formatter_class=help_formatter,
    )

    parser.add_argument("input", type=Path, help="path to encrypted file")
    parser.add_argument("-p", "--passphrase", type=str, help="user passphrase or certificate passphrase")
    parser.add_argument("-c", "--certificate", type=Path, help="user certificate")
    parser.add_argument("-o", "--output", type=Path, required=True, help="path to output directory")

    args = parser.parse_args()

    in_path: Path = args.input.resolve()
    out_path: Path = args.output.resolve()

    if not in_path.exists():
        parser.exit(f"Input file does not exist: {in_path}")

    if not out_path.is_dir():
        parser.exit(f"Output directory does not exist: {out_path}")

    if in_path.parent == out_path:
        parser.exit("Output directory cannot be the same as the input file directory")

    if not args.passphrase and not args.certificate:
        parser.exit("No passphrase or certificate provided")

    segments = find_ad1_files(in_path) if in_path.suffix.lower() == ".ad1" else find_ewf_files(in_path)
    if not segments:
        parser.exit(f"No AD1 or E01 segment files found at: {in_path}")

    with segments[0].open("rb") as fh:
        if not is_adcrypt(fh):
            parser.exit(f"File is not an ADCRYPT container: {segments[0]}")

        adcrypt = ADCrypt(fh)

    try:
        adcrypt.unlock(passphrase=args.passphrase, private_key=args.certificate)
    except (ValueError, TypeError) as e:
        log.exception(e, exc_info=False)  # noqa: TRY401
        parser.exit(1)

    for i, segment in enumerate(segments):
        with segment.open("rb") as fh_in, out_path.joinpath(segment.name).open("wb") as fh_out:
            fh_crypt = adcrypt.wrap(fh_in, index=i)
            log.info("Decrypting segment file %r (%s MB) ..", segment.name, fh_crypt.size // 1024 // 1024)
            shutil.copyfileobj(fh_crypt, fh_out)

    log.info("Finished decrypting file(s), result saved to %s", out_path)



from __future__ import annotations

import hashlib
import hmac
from pathlib import Path
from typing import BinaryIO

from dissect.evidence.adcrypt.c_adcrypt import c_adcrypt
from dissect.evidence.adcrypt.stream import ADCryptStream

try:
    from Crypto.Cipher import AES, PKCS1_v1_5
    from Crypto.Protocol.KDF import PBKDF2
    from Crypto.PublicKey import RSA
    from Crypto.Util import Counter

    HAS_CRYPTO = True
except ImportError:
    HAS_CRYPTO = False


def is_adcrypt(fh: BinaryIO) -> bool:
    """Check if the file handle is an ADCRYPT container.

    Args:
        fh: The file handle to check.
    """
    fh.seek(0)
    return fh.read(8) == c_adcrypt.ADCRYPT_MAGIC.encode()


class ADCrypt:
    """Access Data ADCRYPT encrypted container implementation.

    Not particularly useful on its own, but used by other evidence types such as AD1.
    Pass the first segment file handle to this class, then use :meth:`unlock` to unlock the container,
    and :meth:`wrap` to wrap other segment file handles into decrypting streams.

    References:
        - Reverse engineering ``adencrypt.dll``
        - https://github.com/libyal/libewf/blob/main/documentation/Expert%20Witness%20Compression%20Format%20(EWF).asciidoc#7-ad-encryption
        - https://github.com/log2timeline/plaso/issues/2726#issuecomment-517444736
    """

    def __init__(self, fh: BinaryIO) -> None:
        self.fh = fh
        self.fh.seek(0)

        try:
            self.header: c_adcrypt.Header = c_adcrypt.Header(self.fh)
        except EOFError:
            raise ValueError("File handle is not an ADCRYPT container: Unable to read ADCRYPT header")

        if self.header.magic != c_adcrypt.ADCRYPT_MAGIC.encode():
            raise ValueError(f"File handle is not an ADCRYPT container: Unknown magic {self.header.magic!r}")

        if self.header.version != 1:
            raise ValueError(f"Unsupported ADCRYPT container version {self.header.version!r}")

        self.key: bytes | None = None

    def is_locked(self) -> bool:
        """Return whether the ADCRYPT container is locked."""
        return self.key is None

    def unlock(self, *, passphrase: str | bytes | None = None, private_key: Path | bytes | None = None) -> None:
        """Unlock the ADCRYPT container with a given passphrase or private key.

        Args:
            passphrase: The passphrase to unlock the container.
            private_key: The private key to unlock the container.

        Raises:
            RuntimeError: If required dependencies are missing.
            ValueError: If unlocking failed.
        """
        if not HAS_CRYPTO:
            raise RuntimeError("Missing required dependency 'pycryptodome' for ADCRYPT decryption")

        pkey = adcrypt_kdf(
            passphrase,
            private_key,
            self.header.enc_salt,
            self.header.key_len,
            self.header.pbkdf2_count,
            self.header.hash_algo.name.lower(),
        )

        # Verify the HMAC of EKEY using PKEY + hash algo, comparing with header HMAC
        if hmac.digest(pkey, self.header.enc_key, self.header.hash_algo.name.lower()) != self.header.hmac_enc_key:
            raise ValueError("Unable to unlock: HMAC verification of passphrase failed")

        # Decrypt EKEY using PKEY
        ctr = Counter.new(128, initial_value=0, little_endian=True)
        cipher = AES.new(pkey, AES.MODE_CTR, counter=ctr)
        self.key = cipher.decrypt(self.header.enc_key)

    def wrap(self, fh: BinaryIO, index: int) -> ADCryptStream:
        """Wrap a file handle into an :class:`ADCryptStream` for decryption.

        Args:
            fh: The file handle to wrap.
            index: The segment index.

        Raises:
            ValueError: If the container is not unlocked.
        """
        if self.is_locked():
            raise ValueError("ADCRYPT container is not unlocked")

        return ADCryptStream(fh, self.key, index)


def adcrypt_kdf(
    passphrase: str | bytes | None,
    private_key: Path | bytes | None,
    salt: bytes,
    key_len: int,
    count: int,
    algorithm: str,
) -> bytes:
    """Derive the ADCRYPT decryption key.

    Args:
        passphrase: The passphrase to unlock the container.
        private_key: The private key to unlock the container.
        salt: The salt used for key derivation.
        key_len: The length of the derived key.
        count: The number of iterations for PBKDF2.
        algorithm: The hash algorithm to use.

    Returns:
        The derived key as bytes.
    """
    if isinstance(passphrase, str):
        passphrase = passphrase.encode()

    # If a private key was used, the passphrase is empty.
    passphrase_hash = b""
    if passphrase is not None and private_key is None:
        passphrase_hash = hashlib.new(algorithm, passphrase).digest()

    # If no private key was used, the "encrypted" salt is the plaintext salt as-is.
    derived_salt = salt

    # Decrypt the salt if a private key was provided.
    if private_key is not None:
        rsa_key = RSA.import_key(private_key.read_bytes() if isinstance(private_key, Path) else private_key, passphrase)
        if not (derived_salt := PKCS1_v1_5.new(rsa_key).decrypt(salt, sentinel=None, expected_pt_len=16)):
            raise ValueError("Failed to decrypt salt using provided private key")

    return PBKDF2(passphrase_hash, derived_salt, key_len, count)



# Generated by cstruct-stubgen
from typing import BinaryIO, TypeAlias, overload

import dissect.cstruct as __cs__

class _c_asdf(__cs__.cstruct):
    class FILE_FLAG(__cs__.Flag):
        SHA256 = ...

    class BLOCK_FLAG(__cs__.Flag):
        CRC32 = ...
        COMPRESS = ...

    class header(__cs__.Structure):
        magic: __cs__.CharArray
        flags: _c_asdf.FILE_FLAG
        version: _c_asdf.uint8
        reserved1: __cs__.CharArray
        timestamp: _c_asdf.uint64
        reserved2: __cs__.CharArray
        guid: __cs__.CharArray
        @overload
        def __init__(
            self,
            magic: __cs__.CharArray | None = ...,
            flags: _c_asdf.FILE_FLAG | None = ...,
            version: _c_asdf.uint8 | None = ...,
            reserved1: __cs__.CharArray | None = ...,
            timestamp: _c_asdf.uint64 | None = ...,
            reserved2: __cs__.CharArray | None = ...,
            guid: __cs__.CharArray | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class block(__cs__.Structure):
        magic: __cs__.CharArray
        flags: _c_asdf.BLOCK_FLAG
        idx: _c_asdf.uint8
        reserved: __cs__.CharArray
        offset: _c_asdf.uint64
        size: _c_asdf.uint64
        @overload
        def __init__(
            self,
            magic: __cs__.CharArray | None = ...,
            flags: _c_asdf.BLOCK_FLAG | None = ...,
            idx: _c_asdf.uint8 | None = ...,
            reserved: __cs__.CharArray | None = ...,
            offset: _c_asdf.uint64 | None = ...,
            size: _c_asdf.uint64 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class table_entry(__cs__.Structure):
        flags: _c_asdf.BLOCK_FLAG
        idx: _c_asdf.uint8
        reserved: __cs__.CharArray
        offset: _c_asdf.uint64
        size: _c_asdf.uint64
        file_offset: _c_asdf.uint64
        file_size: _c_asdf.uint64
        @overload
        def __init__(
            self,
            flags: _c_asdf.BLOCK_FLAG | None = ...,
            idx: _c_asdf.uint8 | None = ...,
            reserved: __cs__.CharArray | None = ...,
            offset: _c_asdf.uint64 | None = ...,
            size: _c_asdf.uint64 | None = ...,
            file_offset: _c_asdf.uint64 | None = ...,
            file_size: _c_asdf.uint64 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class footer(__cs__.Structure):
        magic: __cs__.CharArray
        reserved: __cs__.CharArray
        table_offset: _c_asdf.uint64
        sha256: __cs__.CharArray
        @overload
        def __init__(
            self,
            magic: __cs__.CharArray | None = ...,
            reserved: __cs__.CharArray | None = ...,
            table_offset: _c_asdf.uint64 | None = ...,
            sha256: __cs__.CharArray | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

# Technically `c_asdf` is an instance of `_c_asdf`, but then we can't use it in type hints
c_asdf: TypeAlias = _c_asdf



import argparse
import io
import sys
from contextlib import nullcontext
from pathlib import Path
from typing import BinaryIO

from dissect.evidence.asdf import asdf

try:
    from tqdm import tqdm  # type: ignore

    HAS_TQDM = True
except ImportError:
    HAS_TQDM = False


class Progress:
    def __init__(self, size: int):
        self.size = size
        if HAS_TQDM:
            self.t = tqdm(total=size, unit="B", unit_scale=True)

    def update(self, offset: int) -> None:
        if HAS_TQDM:
            self.t.update(offset - self.t.n)
        else:
            sys.stderr.write(f"\r{offset / float(self.size) * 100:0.2f}%")
            sys.stderr.flush()

    def close(self) -> None:
        if HAS_TQDM:
            self.t.close()


def copy_stream(fhin: BinaryIO, fhout: BinaryIO, length: int) -> None:
    n, remain = divmod(length, io.DEFAULT_BUFFER_SIZE)
    fhout.writelines(fhin.read(io.DEFAULT_BUFFER_SIZE) for _ in range(n))
    fhout.write(fhin.read(remain))


def fill_zero(fhout: BinaryIO, length: int) -> None:
    n, remain = divmod(length, io.DEFAULT_BUFFER_SIZE)
    fhout.writelines(b"\x00" * io.DEFAULT_BUFFER_SIZE for _ in range(n))
    fhout.write(b"\x00" * remain)


def main() -> int:
    parser = argparse.ArgumentParser(description="Utility to dump ASDF streams")
    parser.add_argument("file", metavar="ASDF", help="ASDF file to dd")
    parser.add_argument("-w", "--writer", default="-", help="file to write to, default is stdout")
    parser.add_argument("-s", "--stream", type=int, default=0, help="stream index to dump (0-255)")
    parser.add_argument("--no-tqdm", action="store_true", default=False, help="disable tqdm progress bar")
    args = parser.parse_args()

    if args.no_tqdm:
        global HAS_TQDM
        HAS_TQDM = False

    with Path(args.file).open("rb") as fh:
        snapshot = asdf.AsdfSnapshot(fh)

        if args.stream > 255 or not snapshot.contains(args.stream):
            parser.print_help()
            print(file=sys.stderr)

            valid_keys = ", ".join(str(i) for i in snapshot.table)
            parser.exit(1, f"invalid stream index, must be one of {valid_keys}")

        stream = snapshot.open(args.stream)
        progress = Progress(stream.size)

        ctx = nullcontext(getattr(sys.stdout, "buffer", sys.stdout)) if args.writer == "-" else None
        with ctx or Path(args.writer).open("wb") as fhout:
            try:
                prev_offset = 0
                for offset, size, _, data_offset in stream.table:
                    stream.fh.seek(data_offset)
                    if fhout.seekable():
                        fhout.seek(offset)
                    else:
                        fill_zero(fhout, offset - prev_offset)
                    copy_stream(stream.fh, fhout, size)

                    progress.update(offset)
                    prev_offset = offset
            except BrokenPipeError:
                pass
            finally:
                progress.close()

    return 0


if __name__ == "__main__":
    try:
        sys.exit(main())
    except KeyboardInterrupt:
        pass



# Generated by cstruct-stubgen
from typing import BinaryIO, Literal, TypeAlias, overload

import dissect.cstruct as __cs__

class _c_ad1(__cs__.cstruct):
    ADSEGMENTEDFILE_MAGIC: Literal["ADSEGMENTEDFILE\x00"] = ...
    class EntryType(__cs__.Enum):
        File = ...
        Unknown_File = ...
        Deleted = ...
        Directory = ...

    class FileClassType(__cs__.Enum):
        File = ...
        Directory = ...
        ReparsePoint = ...

    class MetaType(__cs__.Enum):
        ItemContentHashes = ...
        FileClass = ...
        FileSize = ...
        PhysicalSize = ...
        Timestamps = ...
        StartCluster = ...
        DateAccessed = ...
        DateCreated = ...
        DateModified = ...
        Encrypted = ...
        Compressed = ...
        ActualFile = ...
        StartSector = ...
        ADSCount = ...
        ShortFilename = ...
        Hidden = ...
        System = ...
        ReadOnly = ...
        Archive = ...
        MFTRecordNumber = ...
        MFTDateChanged = ...
        MFTIsResident = ...
        MFTIsOffline = ...
        MFTIsSparse = ...
        MFTIsTemporary = ...
        MFTOwnerSid = ...
        MFTOwnerName = ...
        MFTGroupSid = ...
        MFTGroupName = ...
        MFTFileDateCreated = ...
        MFTFileDateModified = ...
        MFTFileDateAccessed = ...
        MFTFileDateChanged = ...
        MFTFileSize = ...
        MFTFilePhysicalSize = ...
        IndxFilename = ...
        IndxFileSize = ...
        IndxPhysicalSize = ...
        IndxDateCreated = ...
        IndxDateModified = ...
        IndxDateAccessed = ...
        IndxDateChanged = ...
        AceType = ...
        AceInheritable = ...
        AceSID = ...
        AceName = ...
        AceAccessMask = ...
        AceExecuteFile = ...
        AceReadData = ...
        AceWriteData = ...
        AceAppendData = ...
        AceTraverseFolder = ...
        AceListFolder = ...
        AceCreateFiles = ...
        AceCreateFolders = ...
        AceDeleteChildren = ...
        AceDeleteSelf = ...
        AceReadPermissions = ...
        AceChangePermissions = ...
        AceTakeOwnership = ...
        MD5 = ...
        SHA1 = ...
        ClusterSize = ...
        ClusterCount = ...
        FreeClusterCount = ...
        VolumeSerialNumber = ...
        PosixPermissions = ...

    class SegmentedFileHeader(__cs__.Structure):
        magic: __cs__.CharArray
        unk1: _c_ad1.uint32
        unk2: _c_ad1.uint32
        segment_number: _c_ad1.uint32
        segment_count: _c_ad1.uint32
        segment_size: _c_ad1.uint64
        logical_image_offset: _c_ad1.uint32
        padding: __cs__.CharArray
        @overload
        def __init__(
            self,
            magic: __cs__.CharArray | None = ...,
            unk1: _c_ad1.uint32 | None = ...,
            unk2: _c_ad1.uint32 | None = ...,
            segment_number: _c_ad1.uint32 | None = ...,
            segment_count: _c_ad1.uint32 | None = ...,
            segment_size: _c_ad1.uint64 | None = ...,
            logical_image_offset: _c_ad1.uint32 | None = ...,
            padding: __cs__.CharArray | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class LogicalImageHeader(__cs__.Structure):
        magic: __cs__.CharArray
        version: _c_ad1.uint32
        unk1: _c_ad1.uint32
        chunk_size: _c_ad1.uint32
        metadata_offset: _c_ad1.uint64
        first_file_offset: _c_ad1.uint64
        name_len: _c_ad1.uint32
        unk_magic: __cs__.CharArray
        name_offset: _c_ad1.uint64
        attr_guid_offset: _c_ad1.uint64
        unk2: _c_ad1.uint64
        locs_guid_offset: _c_ad1.uint64
        unk3: _c_ad1.uint64
        name: __cs__.CharArray
        @overload
        def __init__(
            self,
            magic: __cs__.CharArray | None = ...,
            version: _c_ad1.uint32 | None = ...,
            unk1: _c_ad1.uint32 | None = ...,
            chunk_size: _c_ad1.uint32 | None = ...,
            metadata_offset: _c_ad1.uint64 | None = ...,
            first_file_offset: _c_ad1.uint64 | None = ...,
            name_len: _c_ad1.uint32 | None = ...,
            unk_magic: __cs__.CharArray | None = ...,
            name_offset: _c_ad1.uint64 | None = ...,
            attr_guid_offset: _c_ad1.uint64 | None = ...,
            unk2: _c_ad1.uint64 | None = ...,
            locs_guid_offset: _c_ad1.uint64 | None = ...,
            unk3: _c_ad1.uint64 | None = ...,
            name: __cs__.CharArray | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class FileEntry(__cs__.Structure):
        next: _c_ad1.uint64
        child: _c_ad1.uint64
        meta: _c_ad1.uint64
        zlib_meta: _c_ad1.uint64
        size: _c_ad1.uint64
        type: _c_ad1.EntryType
        name_len: _c_ad1.uint32
        name: __cs__.CharArray
        parent_index: _c_ad1.uint64
        @overload
        def __init__(
            self,
            next: _c_ad1.uint64 | None = ...,
            child: _c_ad1.uint64 | None = ...,
            meta: _c_ad1.uint64 | None = ...,
            zlib_meta: _c_ad1.uint64 | None = ...,
            size: _c_ad1.uint64 | None = ...,
            type: _c_ad1.EntryType | None = ...,
            name_len: _c_ad1.uint32 | None = ...,
            name: __cs__.CharArray | None = ...,
            parent_index: _c_ad1.uint64 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class FileEntryChunks(__cs__.Structure):
        num_chunks: _c_ad1.uint64
        chunks: __cs__.Array[_c_ad1.uint64]
        @overload
        def __init__(
            self, num_chunks: _c_ad1.uint64 | None = ..., chunks: __cs__.Array[_c_ad1.uint64] | None = ...
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class FileMeta(__cs__.Structure):
        next: _c_ad1.uint64
        category: _c_ad1.uint32
        type: _c_ad1.MetaType
        len: _c_ad1.uint32
        data: __cs__.CharArray
        @overload
        def __init__(
            self,
            next: _c_ad1.uint64 | None = ...,
            category: _c_ad1.uint32 | None = ...,
            type: _c_ad1.MetaType | None = ...,
            len: _c_ad1.uint32 | None = ...,
            data: __cs__.CharArray | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class Footer(__cs__.Structure):
        unk1: __cs__.CharArray
        @overload
        def __init__(self, unk1: __cs__.CharArray | None = ...): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class ReparsePoint(__cs__.Structure):
        unknown: __cs__.CharArray
        link: __cs__.WcharArray
        @overload
        def __init__(self, unknown: __cs__.CharArray | None = ..., link: __cs__.WcharArray | None = ...): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

# Technically `c_ad1` is an instance of `_c_ad1`, but then we can't use it in type hints
c_ad1: TypeAlias = _c_ad1



from __future__ import annotations

import hashlib
import logging
from typing import TYPE_CHECKING

from dissect.evidence.ad1.ad1 import find_files
from dissect.evidence.tools import adcrypt
from tests._utils import absolute_path

if TYPE_CHECKING:
    from pathlib import Path

    import pytest


def test_adcrypt_passphrase(tmp_path: Path, caplog: pytest.LogCaptureFixture, monkeypatch: pytest.MonkeyPatch) -> None:
    """Test if we can decrypt ADCRYPT AD1 images using the adcrypt tool."""

    with caplog.at_level(logging.DEBUG, adcrypt.log.name), monkeypatch.context() as m:
        m.setattr(
            "sys.argv",
            [
                "adcrypt",
                str(absolute_path("_data/ad1/encrypted-passphrase/encrypted.ad1")),
                "-p",
                "password",
                "-o",
                str(tmp_path),
            ],
        )

        adcrypt.main()

        for i in range(1, 14):
            assert f"Decrypting segment file 'encrypted.ad{i}'" in caplog.text

        assert tmp_path.joinpath("encrypted.ad1").exists()

        ctx = hashlib.sha1()
        for path in find_files(tmp_path.joinpath("encrypted.ad1")):
            ctx.update(path.read_bytes())

        assert ctx.hexdigest() == "3b7449fd09e5803006ce1b3aba5bb4c48c083f12"


def test_adcrypt_certificate(tmp_path: Path, caplog: pytest.LogCaptureFixture, monkeypatch: pytest.MonkeyPatch) -> None:
    """Test if we can decrypt ADCRYPT AD1 images using the adcrypt tool."""

    with caplog.at_level(logging.DEBUG, adcrypt.log.name), monkeypatch.context() as m:
        m.setattr(
            "sys.argv",
            [
                "adcrypt",
                str(absolute_path("_data/ad1/encrypted-certificate/encrypted.ad1")),
                "-c",
                str(absolute_path("_data/ad1/encrypted-certificate/key")),
                "-o",
                str(tmp_path),
            ],
        )

        adcrypt.main()

        for i in range(1, 8):
            assert f"Decrypting segment file 'encrypted.ad{i}'" in caplog.text

        assert tmp_path.joinpath("encrypted.ad1").exists()

        ctx = hashlib.sha1()
        for path in find_files(tmp_path.joinpath("encrypted.ad1")):
            ctx.update(path.read_bytes())

        assert ctx.hexdigest() == "23cdf7c35327d5b24c81ff48b483ae805c27df6a"



from __future__ import annotations

from dissect.evidence.adcrypt.adcrypt import ADCrypt, is_adcrypt
from dissect.evidence.adcrypt.stream import ADCryptStream
from dissect.evidence.exception import Error

__all__ = [
    "ADCrypt",
    "ADCryptStream",
    "Error",
    "is_adcrypt",
]



from __future__ import annotations

import hashlib
from datetime import datetime, timezone
from typing import TYPE_CHECKING, BinaryIO

import pytest

from dissect.evidence.ad1 import ad1
from dissect.evidence.ad1.ad1 import EntryType, find_files
from tests._utils import absolute_path

if TYPE_CHECKING:
    from pathlib import Path


def test_ad1(ad1_basic: BinaryIO) -> None:
    """Test if we can parse a basic non-segmented AD1 file with no file hierarchy."""

    fs = ad1.AD1(ad1_basic)
    assert fs.segment(0).header.magic == b"ADSEGMENTEDFILE\x00"

    assert fs.root.is_dir()
    assert fs.root.listdir() == ["E:"]

    file = fs.get("E:/AD1_test/doc1.txt")
    assert file.is_file()
    assert file.size == 17
    assert file.atime == datetime(2017, 3, 31, 18, 2, 31, 189682, tzinfo=timezone.utc)
    assert file.open().read() == b"Inhoud document 1"


def test_ad1_long(ad1_long: BinaryIO) -> None:
    """Test if we can parse a basic non-segmented AD1 file with long file names."""

    fs = ad1.AD1(ad1_long)

    assert fs.segment(0).header.magic == b"ADSEGMENTEDFILE\x00"
    assert fs.root.is_dir()

    assert [file.name for file in fs.root.children] == ["E:"]

    assert [file.name for file in fs.get("E:").children] == [
        "testdatamap 2 met spaties en een heel stuk langer",
    ]

    assert [file.name for file in fs.get("E:/testdatamap 2 met spaties en een heel stuk langer").iterdir()] == [
        "een lange filenaam 1 met spaties.txt",
        "Een nog langere bestandsnaam met nog meer tekens en 12345.txt",
    ]

    entry = fs.get("E:/testdatamap 2 met spaties en een heel stuk langer").children[0]
    assert entry.name == "een lange filenaam 1 met spaties.txt"
    assert entry.open().read() == (
        b"masdhdslkfjasdfjlksadjflkjsda;lfj\r\nasdflk\r\na;lsdkf\r\n"
        b";lasdklf;lkasd\r\n;lk\r\nfask;ldkf\r\n;lka\r\nsd;lkf\r\n"
        b"asdfasdaflkjsd;lkg;dfshglkdksfhg;ljsdflgjs;dlkkjg'qwjer'pgtoks\r\n"
        b"ddasd'dgkls'dfkjg\r\nsd'g;lkksd'f';gkjsd\r\n[fkgli'erjrg';ksd\r\n"
        b"'g'asldjg';askg\r\nkqe\r\n-["
    )
    md5sum = hashlib.md5(entry.open().read())
    assert md5sum.hexdigest() == entry.md5


def test_ad1_compressed(ad1_compressed: BinaryIO) -> None:
    """Test if we can parse a non-segmented AD1 file with standard zlib compression."""

    fs = ad1.AD1(ad1_compressed)

    assert fs.segment(0).header.magic == b"ADSEGMENTEDFILE\x00"

    assert fs.get("/").listdir() == ["E:"]
    assert fs.get("E:/AD1_test").listdir() == ["doc1.txt", "doc2.txt"]
    assert fs.get("E:/AD1_test/doc1.txt").open().read() == b"Inhoud document 1"


@pytest.mark.parametrize(
    ("path", "expected_files"),
    [
        pytest.param(
            "_data/ad1/pcbje/text-and-pictures.ad1",
            [
                "text-and-pictures.ad1",
                "text-and-pictures.ad2",
                "text-and-pictures.ad3",
                "text-and-pictures.ad4",
            ],
            id="segmented-simple",
        ),
        pytest.param(
            "_data/ad1/encrypted-passphrase/encrypted.ad1",
            [
                "encrypted.ad1",
                "encrypted.ad2",
                "encrypted.ad3",
                "encrypted.ad4",
                "encrypted.ad5",
                "encrypted.ad6",
                "encrypted.ad7",
                "encrypted.ad8",
                "encrypted.ad9",
                "encrypted.ad10",
                "encrypted.ad11",
                "encrypted.ad12",
                "encrypted.ad13",
            ],
            id="segmented-natural-sorting",
        ),
    ],
)
def test_ad1_find_files(path: str, expected_files: list[str]) -> None:
    """Test if we correctly find and order segmented AD1 files and do not find .txt or .csv artifact files."""

    files = find_files(absolute_path(path))
    assert [file.name for file in files] == expected_files


def test_ad1_segmented(ad1_segmented: list[Path]) -> None:
    """Test if we can parse segmented AD1 files.

    References:
        - https://github.com/pcbje/pyad1/tree/master/test_data
    """

    fs = ad1.AD1(ad1_segmented)

    assert len(fs.fh) == 4
    assert fs.segment(0).number == 1
    assert fs.segment(0).count == 4
    assert fs.segment(0).size == 0x200000 - 512

    assert fs.logical_image.version == 4
    assert fs.logical_image.name == "C:\\Users\\pcbje\\Desktop\\Data"

    dir = fs.get("C:/Users/pcbje/Desktop/Data/Pictures")
    assert dir.is_dir()
    assert not dir.is_symlink()
    assert not dir.is_file()
    assert dir.name == "Pictures"
    assert dir.type == EntryType.Directory
    assert dir.size == 0
    assert dir.btime == datetime(2018, 5, 2, 7, 34, 11, 284926, tzinfo=timezone.utc)
    assert dir.mtime == datetime(2018, 5, 2, 7, 42, 39, 841574, tzinfo=timezone.utc)
    assert dir.ctime == datetime(2018, 5, 2, 7, 42, 39, 841574, tzinfo=timezone.utc)
    assert dir.atime == datetime(2018, 5, 2, 7, 46, 59, 164650, tzinfo=timezone.utc)
    assert dir.listdir() == [
        "0-0-581-Hydrangeas.jpg",
        "1-0-858-Chrysanthemum.jpg",
        "2-0-826-Desert.jpg",
        "4-0-757-Jellyfish.jpg",
        "5-0-762-Koala.jpg",
        "6-0-548-Lighthouse.jpg",
        "7-0-759-Penguins.jpg",
    ]

    picture = fs.get("C:/Users/pcbje/Desktop/Data/Pictures/5-0-762-Koala.jpg")
    assert picture.is_file()
    assert not picture.is_dir()
    assert not picture.is_symlink()
    assert picture.btime == datetime(2018, 1, 28, 7, 18, 0, tzinfo=timezone.utc)
    assert picture.mtime == datetime(2018, 5, 2, 7, 42, 34, 287014, tzinfo=timezone.utc)
    assert picture.ctime == datetime(2018, 5, 2, 7, 42, 34, 287014, tzinfo=timezone.utc)
    assert picture.atime == datetime(2018, 5, 2, 7, 42, 35, 611785, tzinfo=timezone.utc)

    buf = picture.open().read()
    assert picture.name == "5-0-762-Koala.jpg"
    assert picture.size == 780831
    assert len(buf) == 780831
    assert picture.sha1 == "9c3dcb1f9185a314ea25d51aed3b5881b32f420c"
    assert hashlib.sha1(buf).hexdigest() == "9c3dcb1f9185a314ea25d51aed3b5881b32f420c"


def test_adcrypt_passphrase(ad1_encrypted_passphrase: list[Path]) -> None:
    """Test if we can decrypt ADCRYPT AD1 images, in this example a segmented AD1 logical image."""
    fs = ad1.AD1(ad1_encrypted_passphrase)

    assert fs.is_adcrypt()
    assert fs.is_locked()

    with pytest.raises(ValueError, match="AD1 container is locked by ADCRYPT"):
        fs.get("/")

    with pytest.raises(ValueError, match="Unable to unlock: HMAC verification of passphrase failed"):
        fs.unlock(passphrase="asdf")

    fs.unlock(passphrase="password")

    assert fs.adcrypt.key.hex() == "9030a43f29689a045e815cf4f0ad82b68850063b414f2797f0897e188f98d7b4"

    assert fs.get("C:/Users/User/Downloads").listdir() == [
        "7z2501-x64.exe",
        "desktop.ini",
        "Exterro_FTK_Imager_(x64)-4.7.3.81.exe",
        "hans-veth-8y--BAFlC9c-unsplash.jpg",
        "marc-olivier-jodoin-tauPAnOIGvE-unsplash.jpg",
        "marek-szturc-8Ou3EZmTMWA-unsplash.jpg",
        "milo-weiler-1AIYdIb3O5M-unsplash.jpg",
    ]

    for file in fs.get("C:/Users/User/Downloads").iterdir():
        buf = file.open().read()
        assert len(buf) == file.size
        assert hashlib.sha1(buf).hexdigest() == file.sha1


def test_adcrypt_certificate(ad1_encrypted_certificate: list[Path]) -> None:
    """Test if we can decrypt ADCRYPT AD1 images, in this example a segmented AD1 logical image."""
    fs = ad1.AD1(ad1_encrypted_certificate)

    assert fs.is_adcrypt()
    assert fs.is_locked()

    with pytest.raises(ValueError, match="AD1 container is locked by ADCRYPT"):
        fs.get("/")

    with pytest.raises(ValueError, match="Unable to unlock: HMAC verification of passphrase failed"):
        fs.unlock(passphrase="asdf")

    fs.unlock(private_key=absolute_path("_data/ad1/encrypted-certificate/key"))

    assert fs.adcrypt.key.hex() == "6cc0a9f94f944381cc51be474e5da6178059324bb457a87e0035b80f80ff9d4b"

    assert fs.get("C:/Users/User/Downloads").listdir() == [
        "desktop.ini",
        "hans-veth-8y--BAFlC9c-unsplash.jpg",
        "key.pem",
        "marc-olivier-jodoin-tauPAnOIGvE-unsplash.jpg",
        "marek-szturc-8Ou3EZmTMWA-unsplash.jpg",
        "milo-weiler-1AIYdIb3O5M-unsplash.jpg",
        "programs",
    ]

    for file in fs.get("C:/Users/User/Downloads").iterdir():
        if file.is_dir():
            continue

        buf = file.open().read()
        assert len(buf) == file.size
        assert hashlib.sha1(buf).hexdigest() == file.sha1

    assert fs.get("C:/Users/User/Downloads/programs").listdir() == [
        "7z2501-x64.exe",
        "Exterro_FTK_Imager_(x64)-4.7.3.81.exe",
    ]


def test_ad1_segment_lru(ad1_segmented: list[Path], monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(ad1, "MAX_OPEN_SEGMENTS", 2)

    fs = ad1.AD1(ad1_segmented)
    assert fs._segment_lru == [3, 0]

    fs.segment(2)
    assert fs._segment_lru == [0, 2]

    fs.segment(1)
    assert fs._segment_lru == [2, 1]

    picture = fs.get("C:/Users/pcbje/Desktop/Data/Pictures/5-0-762-Koala.jpg")
    assert hashlib.sha1(picture.open().read()).hexdigest() == "9c3dcb1f9185a314ea25d51aed3b5881b32f420c"



from __future__ import annotations

from dissect.cstruct import cstruct

ad1_def = """
enum EntryType : uint32 {
    File                    = 0x0,
    Unknown_File            = 0x1,
    Deleted                 = 0x2,
    Directory               = 0x5,
};

enum FileClassType : uint32 {
    File                    = 0x1,          // b"1"
    Directory               = 0x3,          // b"3"
    ReparsePoint            = 0x3131,       // b"11"
};

enum MetaType : uint32 {
    // Generic attributes
    ItemContentHashes       = 0x1,
    FileClass               = 0x2,
    FileSize                = 0x3,
    PhysicalSize            = 0x4,
    Timestamps              = 0x5,
    StartCluster            = 0x6,
    DateAccessed            = 0x7,
    DateCreated             = 0x8,
    DateModified            = 0x9,
    // .. 0xa, 0xb, 0xc ..
    Encrypted               = 0xd,
    Compressed              = 0xe,
    // .. 0xf ..
    ActualFile              = 0x1e,
    StartSector             = 0x1f,
    ADSCount                = 0x24,         // Alternate Data Stream Count

    // DOS attributes
    ShortFilename           = 0x1001,
    Hidden                  = 0x1002,
    System                  = 0x1003,
    ReadOnly                = 0x1004,
    Archive                 = 0x1005,

    // NTFS attributes
    MFTRecordNumber         = 0xa001,
    MFTDateChanged          = 0xa002,       // Specifies the MFT record change timestamp of the file.
    MFTIsResident           = 0xa003,
    MFTIsOffline            = 0xa004,
    MFTIsSparse             = 0xa005,
    MFTIsTemporary          = 0xa006,
    MFTOwnerSid             = 0xa007,
    MFTOwnerName            = 0xa008,
    MFTGroupSid             = 0xa009,
    MFTGroupName            = 0xa00a,

    MFTFileDateCreated      = 0xa01c,       // According to the filename attribute in the MFT.
    MFTFileDateModified     = 0xa01d,
    MFTFileDateAccessed     = 0xa01e,
    MFTFileDateChanged      = 0xa01f,
    MFTFileSize             = 0xa020,
    MFTFilePhysicalSize     = 0xa021,

    // 8.3 MFT Filename
    // 0xa022,
    // 0xa023,
    // 0xa024,
    // 0xa025,
    // 0xa026,
    // 0xa027,

    IndxFilename            = 0xa028,       // According to the filename attribute in the $I30 INDX.
    IndxFileSize            = 0xa029,
    IndxPhysicalSize        = 0xa02a,
    IndxDateCreated         = 0xa02b,
    IndxDateModified        = 0xa02c,
    IndxDateAccessed        = 0xa02d,
    IndxDateChanged         = 0xa02e,

    // 8.3 INDX
    // 0xa02f, 0xa030, 0xa031, 0xa032, 0xa033, 0xa034, 0xa035

    // NTFS Access Control Entry (0)
    AceType                 = 0x1000001,
    AceInheritable          = 0x1000004,
    AceSID                  = 0x1000005,    // The Security ID of the user or group this ACE applies to.
    AceName                 = 0x1000006,    // The name of the user or roup this ACE applies to.
    AceAccessMask           = 0x1000007,    // Raw bitmask specifying the actions this ACE controls.
    AceExecuteFile          = 0x1000008,
    AceReadData             = 0x1000009,
    AceWriteData            = 0x100000a,
    AceAppendData           = 0x100000b,
    AceTraverseFolder       = 0x100000c,
    AceListFolder           = 0x100000d,
    AceCreateFiles          = 0x100000e,
    AceCreateFolders        = 0x100000f,
    AceDeleteChildren       = 0x1000010,
    AceDeleteSelf           = 0x1000011,
    AceReadPermissions      = 0x1000012,
    AceChangePermissions    = 0x1000013,
    AceTakeOwnership        = 0x1000014,
    // .. 0x10010XX - 0x10060XX ..

    // Verification hashes
    MD5                     = 0x5001,
    SHA1                    = 0x5002,

    // TODO: Clean up
    ClusterSize             = 0x9001,
    ClusterCount            = 0x9002,
    FreeClusterCount        = 0x9003,
    VolumeSerialNumber      = 0x9006,
    PosixPermissions        = 0x2001,
};

#define ADSEGMENTEDFILE_MAGIC ADSEGMENTEDFILE\00

typedef struct {
    char        magic[16];                  // b"ADSEGMENTEDFILE" + padding
    uint32      unk1;                       // 0x01
    uint32      unk2;                       // 0x02
    uint32      segment_number;             // segment number starts at 0x01
    uint32      segment_count;              // number of segments
    uint64      segment_size;               // off by 512 bytes
    uint32      logical_image_offset;
    char        padding[468];               // 0x00
} SegmentedFileHeader;

typedef struct {
    char        magic[16];                  // b"ADLOGICALIMAGE" + padding
    uint32      version;                    // 0x03 or 0x04
    uint32      unk1;                       // 0x01
    uint32      chunk_size;                 // zlib chunk size (uint64?)
    uint64      metadata_offset;
    uint64      first_file_offset;
    uint32      name_len;

    // ADv4 (offset 48 contains name[name_len] in ADv3)
    char        unk_magic[4];               // b"AD" + (2 * 0x00)
    uint64      name_offset;                // 0x5c
    uint64      attr_guid_offset;
    uint64      unk2;                       // 0x00
    uint64      locs_guid_offset;
    uint64      unk3;                       // 0x00
    // END ADv4

    char        name[name_len];
} LogicalImageHeader;

typedef struct {
    uint64      next;                       // Next FileEntry in same hierarchy level
    uint64      child;                      // Next FileEntry within this dir, 0x00 if file
    uint64      meta;                       // Offset of first FileMeta entry
    uint64      zlib_meta;                  // Offset of zlib chunk metadata
    uint64      size;                       // Decompressed file size, 0x00 if no data
    EntryType   type;                       // 0x00 = file, 0x05 = directory
    uint32      name_len;
    char        name[name_len];
    uint64      parent_index;               // Parent folder index, 0x00 if at root
} FileEntry;

typedef struct {
    uint64      num_chunks;                 // only if FileEntry.size != 0x00
    uint64      chunks[num_chunks];
} FileEntryChunks;

typedef struct {
    uint64      next;
    uint32      category;
    MetaType    type;
    uint32      len;
    char        data[len];
} FileMeta;

typedef struct {
    char        unk1[352];                  // version 4 only
} Footer;

typedef struct {
    CHAR        unknown[16];
    WCHAR       link[EOF];
} ReparsePoint;
"""

c_ad1 = cstruct().load(ad1_def)



from __future__ import annotations

from dissect.cstruct import cstruct

ewf_def = """
enum MediaType : uint8 {
    Removable   = 0x00,
    Fixed       = 0x01,
    Optical     = 0x03,
    Logical     = 0x0e,
    RAM         = 0x10
};

enum MediaFlags : uint8 {
    Image       = 0x01,
    Physical    = 0x02,
    Fastbloc    = 0x04,
    Tablaeu     = 0x08
};

enum CompressionLevel : uint8 {
    None        = 0x00,
    Good        = 0x01,
    Best        = 0x02
};

typedef struct {
    char        signature[8];
    uint8       fields_start;
    uint16      segment_number;
    uint16      fields_end;
} EWFHeader;

typedef struct {
    char    type[16];
    uint64  next;
    uint64  size;
    uint8   pad[40];
    uint32  checksum;
} EWFSectionDescriptor;

typedef struct {
    uint32  reserved_1;
    uint32  chunk_count;
    uint32  sector_count;
    uint32  sector_size;
    uint32  total_sector_count;
    uint8   reserved[20];
    uint8   pad[45];
    char    signature[5];
    uint32  checksum;
} EWFVolumeSectionSpec;

typedef struct {
    MediaType           media_type;
    uint8               reserved_1[3];
    uint32              chunk_count;
    uint32              sector_count;
    uint32              sector_size;
    uint64              total_sector_count;
    uint32              num_cylinders;
    uint32              num_heads;
    uint32              num_sectors;
    uint8               media_flags;
    uint8               unknown_1[3];
    uint32              palm_start_sector;
    uint32              unknown_2;
    uint32              smart_start_sector;
    CompressionLevel    compression_level;
    uint8               unknown_3[3];
    uint32              error_granularity;
    uint32              unknown_4;
    uint8               uuid[16];
    uint8               pad[963];
    char                signature[5];
    uint32              checksum;
} EWFVolumeSection;

typedef struct {
    uint32  num_entries;
    uint32  _;
    uint64  base_offset;
    uint32  _;
    uint32  checksum;
    uint32  entries[num_entries];
} EWFTableSection;
"""

c_ewf = cstruct().load(ewf_def)



from __future__ import annotations

import errno
import os
import sys
from functools import wraps
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable


def catch_sigpipe(func: Callable) -> Callable:
    """Catches ``KeyboardInterrupt`` and ``BrokenPipeError`` (``OSError 22`` on Windows)."""

    @wraps(func)
    def wrapper(*args, **kwargs) -> int:
        try:
            return func(*args, **kwargs)
        except KeyboardInterrupt:
            print("Aborted!", file=sys.stderr)
            return 1
        except OSError as e:
            # Only catch BrokenPipeError or OSError 22
            if e.errno in (errno.EPIPE, errno.EINVAL):
                devnull = os.open(os.devnull, os.O_WRONLY)
                os.dup2(devnull, sys.stdout.fileno())
                return 1
            # Raise other exceptions
            raise

    return wrapper



# Acquire Snapshot Data Format

from __future__ import annotations

import gzip
import io
import shutil
import tarfile
import uuid
from bisect import bisect_right
from collections import defaultdict
from typing import TYPE_CHECKING, BinaryIO

from dissect.util import ts
from dissect.util.stream import AlignedStream, RangeStream

from dissect.evidence.asdf.c_asdf import c_asdf
from dissect.evidence.asdf.stream import CompressedStream, Crc32Stream, HashedStream
from dissect.evidence.exception import (
    InvalidBlock,
    InvalidSnapshot,
    UnsupportedVersion,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator

SnapshotTableEntry = tuple[int, int, int, int]

VERSION = 1
DEFAULT_BLOCK_SIZE = 4096
MAX_BLOCK_TABLE_SIZE = 2**32

MAX_IDX = 253
IDX_MEMORY = 254
IDX_METADATA = 255
RESERVED_IDX = [IDX_METADATA, IDX_MEMORY]

FILE_MAGIC = b"ASDF"
BLOCK_MAGIC = b"BL\xa5\xdf"
FOOTER_MAGIC = b"FT\xa5\xdf"
SPARSE_BYTES = b"\xa5\xdf"


class AsdfWriter(io.RawIOBase):
    """ASDF file writer.

    Current limitations:
        - Maximum source disk size is ~16EiB
        - Maximum number of disks is 254

    Some things are currently hardcoded (like SHA256), although they may
    become variable in the future.

    Args:
        fh: File-like object to write to.
        guid: Unique identifier. Used to link images to writers.
        compress: Write gzip compressed file.
        block_crc: Flag to store a CRC32 after each block.
    """

    def __init__(
        self,
        fh: BinaryIO,
        guid: uuid.UUID | None = None,
        compress: bool = False,
        block_crc: bool = True,
    ):
        self._fh = fh
        self.fh = self._fh

        if compress:
            self.fh = gzip.GzipFile(fileobj=self.fh, mode="wb")

        self.fh = HashedStream(self.fh)
        self.guid = guid or uuid.uuid4()

        # Options
        self.block_crc = block_crc
        self.block_compress = False  # Disabled for now

        self._table = defaultdict(list)
        self._table_lookup = defaultdict(list)
        self._table_offset = 0

        self._meta_buf = io.BytesIO()
        self._meta_tar = tarfile.open(fileobj=self._meta_buf, mode="w")  # noqa: SIM115

        self._write_header()

    def add_metadata_file(self, path: str, fh: BinaryIO, size: int | None = None) -> None:
        """Add a file to the metadata stream.

        Args:
            path: The path in the metadata tar to write to.
            fh: The file-like object to write.
            size: Optional size to write.
        """
        info = self._meta_tar.tarinfo()
        info.name = path
        info.uname = "root"
        info.gname = "root"

        if size is None and fh.seekable():
            fh.seek(0, io.SEEK_END)
            size = fh.tell()

        info.size = size or 0

        fh.seek(0)
        self._meta_tar.addfile(info, fh)

    def add_bytes(self, data: bytes, idx: int = 0, base: int = 0) -> None:
        """Add some bytes into this snapshot.

        Convenience method for adding some bytes at a specific offset.

        Args:
            data: The bytes to copy.
            idx: The stream index.
            base: The base offset.
        """
        self._write_block(io.BytesIO(data), 0, len(data), idx=idx, base=base)

    def copy_bytes(self, source: BinaryIO, offset: int, num_bytes: int, idx: int = 0, base: int = 0) -> None:
        """Copy some bytes from the source file-like object into this snapshot.

        Often the source will be a volume on a disk, which is usually represented
        as a relative stream. If this is the case, use the ``base`` argument to
        indicate what the byte offset of the source is, relative to the start
        of the disk. The ``offset`` argument is always the offset in the
        source, so that is not affected.

        Args:
            source: The source file-like object to copy the bytes from.
            offset: The byte offset into the source to start copying bytes from.
            num_bytes: The amount of bytes to copy.
            idx: The stream index, if copying from multiple disks.
            base: The base offset, if the source is a relative stream from e.g. a disk.
        """
        self._write_block(source, offset, num_bytes, idx=idx, base=base)

    def copy_block(
        self,
        source: BinaryIO,
        offset: int,
        num_blocks: int,
        block_size: int | None = None,
        idx: int = 0,
        base: int = 0,
    ) -> None:
        """Copy some blocks in the given block size into this snapshot.

        If no block size is given, the ASDF native block size is used.
        This is really just a convenience method that does the block multiplication
        before calling ``copy_bytes``.

        Args:
            source: The source file-like object to copy the blocks from.
            offset: The byte offset into the source to start copying blocks from.
            num_blocks: The amount of blocks to copy.
            block_size: The size of each block.
            idx: The stream index, if copying from multiple disks.
            base: The base offset, if the source is a relative stream from e.g. a disk.
        """
        block_size = block_size or DEFAULT_BLOCK_SIZE
        return self.copy_bytes(source, offset, num_blocks * block_size, idx, base)

    def copy_runlist(
        self,
        source: BinaryIO,
        runlist: list[tuple[int | None, int]],
        runlist_block_size: int,
        idx: int = 0,
        base: int = 0,
    ) -> None:
        """Copy a runlist of blocks in the given block size into this snapshot.

        A runlist must be a list of tuples, where:
            (block_offset, num_blocks)

        This is really just a convenience method that does the runlist iteration
        and block multiplication before calling `copy_bytes`.

        Args:
            source: The source file-like object to copy the blocks from.
            runlist: The runlist that describes the blocks.
            runlist_block_size: The size of each block.
            idx: The stream index, if copying from multiple disks.
            base: The base offset, if the source is a relative stream from e.g. a disk.
        """
        for run_offset, run_length in runlist:
            # If run_offset is None, it's a sparse run
            if run_offset is None:
                continue

            # Save a function call by directly calling copy_bytes instead of copy_block.
            self.copy_bytes(source, run_offset * runlist_block_size, run_length * runlist_block_size, idx, base)

    def close(self) -> None:
        """Close the ASDF file.

        Writes the block table and footer, then closes the destination file-like object.
        """
        super().close()
        self._write_meta()
        if self._table:
            self._write_table()
        self._write_footer()
        self.fh.close()

    def _write_header(self) -> None:
        """Write the ASDF header to the destination file-like object."""
        header = c_asdf.header(
            magic=FILE_MAGIC,
            flags=c_asdf.FILE_FLAG.SHA256,  # Currently the only option
            version=VERSION,
            timestamp=ts.unix_now(),
            guid=self.guid.bytes_le,
        )
        header.write(self.fh)

    def _write_block(self, source: BinaryIO, offset: int, size: int, idx: int = 0, base: int = 0) -> None:
        """Write an ASDF block to the destination file-like object.

        Args:
            source: The source file-like object to copy the bytes from.
            offset: The byte offset of the copied block.
            size: The size of the copied block in bytes.
            idx: The stream index, if copying from multiple disks.
            base: The base offset, if the source is a relative stream from a disk.
        """
        absolute_offset = base + offset

        lookup_table = self._table_lookup[idx]
        entry_table = self._table[idx]

        table_idx, absolute_offset, size = _table_fit(
            absolute_offset, size, entry_table, lookup_table, lambda e: (e[2], e[3])
        )

        if table_idx is None:
            return

        offset = absolute_offset - base

        # Setup the block flags and block writer
        flags = 0
        outfh = self.fh
        if self.block_crc:
            outfh = Crc32Stream(outfh)
            flags |= c_asdf.BLOCK_FLAG.CRC32
        if self.block_compress:
            outfh = CompressedStream(outfh)
            flags |= c_asdf.BLOCK_FLAG.COMPRESS

        block_offset = self.fh.tell()  # Block header location
        block = c_asdf.block(
            magic=BLOCK_MAGIC,
            flags=flags,
            idx=idx,
            offset=absolute_offset,
            size=size,
        )
        block.write(self.fh)
        data_offset = self.fh.tell()  # Block data location

        source_stream = RangeStream(source, offset, size)
        shutil.copyfileobj(source_stream, outfh)
        # This writes any remaining data or footer for each block writer
        outfh.finalize()

        data_size = self.fh.tell() - data_offset

        lookup_table.insert(table_idx, absolute_offset)
        entry_table.insert(table_idx, (flags, idx, absolute_offset, size, block_offset, data_size))

    def _write_meta(self) -> None:
        """Write the metadata tar to the destination file-like object."""
        self._meta_tar.close()

        size = self._meta_buf.tell()
        self._meta_buf.seek(0)
        self.copy_bytes(self._meta_buf, 0, size, idx=IDX_METADATA)

    def _write_table(self) -> None:
        """Write the ASDF block table to the destination file-like object."""
        self._table_offset = self.fh.tell()
        for stream_table in self._table.values():
            for flags, idx, offset, size, file_offset, file_size in stream_table:
                table_entry = c_asdf.table_entry(
                    flags=flags,
                    idx=idx,
                    offset=offset,
                    size=size,
                    file_offset=file_offset,
                    file_size=file_size,
                )
                table_entry.write(self.fh)

    def _write_footer(self) -> None:
        """Write the ASDF footer to the destination file-like object."""
        footer = c_asdf.footer(
            magic=FOOTER_MAGIC,
            table_offset=self._table_offset,
            sha256=self.fh.digest(),
        )
        footer.write(self.fh)


class AsdfSnapshot:
    """ASDF file reader.

    Args:
        fh: File-like object to read the ASDF file from.
    """

    def __init__(self, fh: BinaryIO, recover: bool = False):
        self.fh = fh
        self.header = c_asdf.header(fh)
        if self.header.magic != FILE_MAGIC:
            raise InvalidSnapshot("invalid header magic")

        if self.header.version > VERSION:
            raise UnsupportedVersion("higher version")

        self.timestamp = ts.from_unix(self.header.timestamp)
        self.guid = uuid.UUID(bytes_le=self.header.guid)

        self.table: dict[list[SnapshotTableEntry]] = defaultdict(list)
        self._table_lookup: dict[list[int]] = defaultdict(list)

        footer_offset = self.fh.seek(-len(c_asdf.footer), io.SEEK_END)

        self.footer = c_asdf.footer(self.fh)
        if self.footer.magic != FOOTER_MAGIC:
            if recover:
                self._recover_block_table()
            else:
                raise InvalidSnapshot("invalid footer magic")
        else:
            self._parse_block_table(
                self.footer.table_offset,
                (footer_offset - self.footer.table_offset) // len(c_asdf.table_entry),
            )

        self.metadata = Metadata(self)

    def _parse_block_table(self, offset: int, count: int) -> None:
        """Parse the block table, getting rid of overlapping blocks."""
        self.fh.seek(offset)
        table_data = io.BytesIO(self.fh.read(count * len(c_asdf.table_entry)))

        for _ in range(count):
            entry = c_asdf.table_entry(table_data)
            self._table_insert(entry.idx, entry.offset, entry.size, entry.file_offset)

    def _recover_block_table(self) -> None:
        self.fh.seek(len(c_asdf.header))
        for block, file_offset in scrape_blocks(self.fh):
            self._table_insert(block.idx, block.offset, block.size, file_offset)

    def _table_insert(self, idx: int, offset: int, size: int, file_offset: int) -> None:
        stream_idx = idx
        entry_data_offset = file_offset + len(c_asdf.block)

        lookup_table = self._table_lookup[stream_idx]
        entry_table = self.table[stream_idx]

        table_idx, entry_offset, entry_size = _table_fit(
            offset, size, entry_table, lookup_table, lambda e: (e[0], e[1])
        )

        if table_idx is None:
            return

        entry_data_offset += entry_offset - offset

        lookup_table.insert(table_idx, entry_offset)
        entry_table.insert(
            table_idx,
            (
                entry_offset,
                entry_size,
                file_offset,
                entry_data_offset,
            ),
        )

    def contains(self, idx: int) -> bool:
        """Check whether this file contains the given stream index.

        Args:
            idx: The stream to check.
        """
        return idx in self.table

    def open(self, idx: int) -> AsdfStream:
        """Open a specific stream in the file.

        Args:
            idx: The stream to open.
        """
        if idx not in self.table:
            raise IndexError(f"invalid stream idx: {idx}")
        return AsdfStream(self, idx)

    def streams(self) -> Iterator[AsdfStream]:
        """Iterate over all streams in the file."""
        for i in sorted(self.table.keys()):
            yield self.open(i)

    def disks(self) -> Iterator[AsdfStream]:
        """Iterate over all non-reserved streams in the file."""
        for i in sorted(self.table.keys()):
            if i in RESERVED_IDX:
                continue
            yield self.open(i)


class Metadata:
    """ASDF metadata reader.

    Thin wrapper around ``tarfile``.

    Args:
        asdf: The :class:`AsdfSnapshot` to open the metadata of.
    """

    def __init__(self, asdf: AsdfSnapshot):
        self.tar = None
        if IDX_METADATA in asdf.table:
            self.tar = tarfile.open(fileobj=asdf.open(IDX_METADATA), mode="r")  # noqa: SIM115

    def names(self) -> list[str]:
        """Return all metadata file entries."""
        return self.tar.getnames() if self.tar else []

    def members(self) -> list[tarfile.TarInfo]:
        """Return all metadata :class:`tarfile.TarInfo` entries."""
        return self.tar.getmembers() if self.tar else []

    def open(self, path: str) -> BinaryIO:
        """Open a metadata entry and return a binary file-like object."""
        if self.tar:
            return self.tar.extractfile(path)
        raise KeyError(f"filename '{path}' not found")


class AsdfStream(AlignedStream):
    """ASDF stream from a snapshot.

    Args:
        asdf: :class:`AsdfSnapshot` parent object.
        idx: Stream index in the :class:`AsdfSnapshot`.
    """

    def __init__(self, asdf: AsdfSnapshot, idx: int):
        self.fh = asdf.fh
        self.asdf = asdf
        self.idx = idx
        self.table = asdf.table[idx]
        self._table_lookup = asdf._table_lookup[idx]

        # We don't actually know the size of the source disk
        # Doesn't really matter though, just take the last run offset + size
        size = self.table[-1][0] + self.table[-1][1]
        super().__init__(size)

    def _read(self, offset: int, length: int) -> bytes:
        result = []

        size = self.size
        run_idx = bisect_right(self._table_lookup, offset) - 1
        runlist_len = len(self.table)

        while length > 0 and run_idx < runlist_len:
            run_start, run_size, run_file_offset, run_data_offset = self.table[run_idx]
            run_end = run_start + run_size

            if run_idx + 1 < runlist_len:
                next_run_start, _, _, _ = self.table[run_idx + 1]
            else:
                next_run_start = None

            if run_idx < 0:
                # Missing first block
                if next_run_start is None:
                    break

                sparse_remaining = next_run_start - offset

                read_count = min(size - offset, min(sparse_remaining, length))
                result.append(SPARSE_BYTES * (read_count // len(SPARSE_BYTES)))

                # Proceed to next run
                run_idx += 1
            elif run_end <= offset:
                # Start outside of run bounds
                if next_run_start is None:
                    # No next run to sparse read to
                    break

                sparse_size = next_run_start - run_end
                sparse_pos = offset - run_end
                sparse_remaining = sparse_size - sparse_pos

                read_count = min(size - offset, min(sparse_remaining, length))
                result.append(SPARSE_BYTES * (read_count // len(SPARSE_BYTES)))

                # Proceed to next run
                run_idx += 1
            elif offset < run_start:
                # Previous run consumed, and next run is far away
                sparse_remaining = run_start - offset
                read_count = min(size - offset, min(sparse_remaining, length))
                result.append(SPARSE_BYTES * (read_count // len(SPARSE_BYTES)))

                # Don't proceed to next run, next loop iteration we'll be within the current run
            else:
                # We're in a run with data
                run_pos = offset - run_start
                run_remaining = run_size - run_pos
                read_count = min(size - offset, min(run_remaining, length))

                self.fh.seek(run_file_offset)
                if self.fh.read(4) != BLOCK_MAGIC:
                    raise InvalidBlock("invalid block magic")

                # Skip over block header
                self.fh.seek(run_data_offset + run_pos)
                result.append(self.fh.read(read_count))

                # Proceed to next run
                run_idx += 1

            offset += read_count
            length -= read_count

        return b"".join(result)


def scrape_blocks(fh: BinaryIO, buffer_size: int = io.DEFAULT_BUFFER_SIZE) -> Iterator[c_asdf.block, int]:
    """Scrape for block headers in ``fh`` and yield parsed block headers and their offset.

    Args:
        fh: The file-like object to scrape for block headers.
        buffer_size: The buffer size to use when scraping.
    """
    # If BLOCK_MAGIC is not found in the buffer (4 bytes), it's possible that part of it (up to 3 bytes) is in there
    # Keep an overlap buffer of 3 bytes that we prepend to the current buffer so we can also find these partial needles
    overlap_len = len(BLOCK_MAGIC) - 1
    overlap = b"\x00" * overlap_len
    while True:
        pos = fh.tell()
        buf = fh.read(buffer_size)
        if not buf:
            break

        data = overlap + buf
        needle_pos = -1
        while True:
            needle_pos = data.find(BLOCK_MAGIC, needle_pos + 1)
            if needle_pos == -1:
                break

            offset = pos + needle_pos - overlap_len

            fh.seek(offset)
            block_buf = fh.read(len(c_asdf.block))

            # Some sanity checks that this is actually a block header
            if block_buf[4] not in c_asdf.BLOCK_FLAG:
                continue

            if block_buf[6:8] != b"\x00\x00":
                continue

            block_entry = c_asdf.block(block_buf)
            yield block_entry, offset

        # Keep the last 3 bytes as overlap
        overlap = data[-overlap_len:]
        # Consumer may seek the fh, so seek it back to where we were
        fh.seek(pos + len(buf))


def _table_fit(
    entry_offset: int, entry_size: int, entry_table: list, lookup_table: list, getentry: Callable
) -> tuple[int, int, int]:
    """Calculate where to insert an entry with the given offset and size into the entry table.

    Moves or shrinks the entry to prevent block overlap, and remove any overlapping blocks.

    Args:
        entry_offset: The entry offset to calculate the insert for.
        entry_size: The entry size to calculate the insert for.
        entry_table: The entry table to insert into or remove entries from.
        lookup_table: The lookup table for the entry_table.
        getentry: A callable to return the ``(offset, size)`` tuple from an entry.

    Returns:
        A tuple of the table index to insert into, an adjusted entry offset and an adjusted entry size.
    """
    entry_end = entry_offset + entry_size

    prev_end = None
    next_start = None
    next_end = None

    table_idx = bisect_right(lookup_table, entry_offset)
    if table_idx > 0:
        prev_start, prev_size = getentry(entry_table[table_idx - 1])
        prev_end = prev_start + prev_size
    if table_idx < len(lookup_table):
        next_start, next_size = getentry(entry_table[table_idx])
        next_end = next_start + next_size

    if prev_end and prev_end >= entry_end:
        # This block is fully contained in the previous block
        return None, None, None

    if prev_end and prev_end > entry_offset:
        # The start of this block overlaps with the previous, so shrink this block
        entry_offset = prev_end

    # We may completely overlap one or more next entries
    while next_end and next_end <= entry_end:
        lookup_table.pop(table_idx)
        entry_table.pop(table_idx)

        if table_idx < len(lookup_table):
            next_start, next_size = getentry(entry_table[table_idx])
            next_end = next_start + next_size
        else:
            next_start, next_end = None, None

    if next_start and next_start < entry_end < next_end:
        # The next block overlaps with this block, so shrink this block
        entry_end = next_start

    if entry_offset >= entry_end:
        # Shouldn't be possible to go beyond the end, but we may end up with a 0 sized block
        return None, None, None

    return table_idx, entry_offset, entry_end - entry_offset



from __future__ import annotations

import zlib
from bisect import bisect_right
from typing import TYPE_CHECKING

from dissect.util.stream import AlignedStream

from dissect.evidence.ad1.c_ad1 import c_ad1

if TYPE_CHECKING:
    from dissect.evidence.ad1.ad1 import AD1, FileEntry


class AD1Stream(AlignedStream):
    """Provide a stitched stream over all AD1 segments."""

    def __init__(self, ad1: AD1):
        self.ad1 = ad1
        super().__init__(self.ad1.size)

    def _read(self, offset: int, length: int) -> bytes:
        result = []

        idx = bisect_right(self.ad1._segment_offsets, offset)
        while length > 0:
            if idx > len(self.ad1._segment_offsets) - 1:
                break

            segment = self.ad1.segment(idx)
            segment_offset = 0 if idx == 0 else self.ad1._segment_offsets[idx - 1]
            offset_in_segment = offset - segment_offset
            read_size = min(length, segment.size - offset_in_segment)

            segment.fh.seek(512 + offset_in_segment)  # Skip segment header
            result.append(segment.fh.read(read_size))

            offset += read_size
            length -= read_size
            idx += 1

        return b"".join(result)


class FileStream(AlignedStream):
    """Custom stream implementation for AD1 :class:`FileEntry` file contents."""

    def __init__(self, entry: FileEntry):
        self.entry = entry
        self.stream = self.entry.ad1.stream
        self.chunk_size = self.entry.ad1.logical_image.chunk_size

        self.stream.seek(self.entry.entry.zlib_meta)
        self.chunks = [*c_ad1.FileEntryChunks(self.stream).chunks, self.entry.entry.meta]

        super().__init__(self.entry.size, self.chunk_size)

    def _read(self, offset: int, length: int) -> bytes:
        result = []

        chunk, offset_in_chunk = divmod(offset, self.chunk_size)
        chunk_count = (length + self.chunk_size - 1) // self.chunk_size
        chunk_offsets = self.chunks[chunk : chunk + chunk_count + 1]

        for i, chunk_offset in enumerate(chunk_offsets[:-1]):
            compressed_chunk_size = chunk_offsets[i + 1] - chunk_offset

            self.stream.seek(chunk_offset)
            buf = zlib.decompress(self.stream.read(compressed_chunk_size))

            read_size = min(length, self.chunk_size - offset_in_chunk)
            result.append(buf[offset_in_chunk : offset_in_chunk + read_size])

            offset += read_size
            length -= read_size
            offset_in_chunk = 0

        return b"".join(result)



from __future__ import annotations

from dissect.evidence.ewf.c_ewf import c_ewf
from dissect.evidence.ewf.ewf import (
    EWF,
    EWFError,
    EWFStream,
    HeaderSection,
    SectionDescriptor,
    Segment,
    TableSection,
    VolumeSection,
    find_files,
)

__all__ = [
    "EWF",
    "EWFError",
    "EWFStream",
    "HeaderSection",
    "SectionDescriptor",
    "Segment",
    "TableSection",
    "VolumeSection",
    "c_ewf",
    "find_files",
]



from __future__ import annotations

from typing import TYPE_CHECKING, BinaryIO
from unittest.mock import MagicMock, patch

from dissect.evidence.ewf import ewf

if TYPE_CHECKING:
    import pytest


def test_ewf(ewf_data: BinaryIO) -> None:
    e = ewf.EWF(ewf_data)

    assert e.size == 4097
    assert e.open().read(4097) == (b"\xde\xad\xbe\xef" * 1024) + b"\n"


@patch("dissect.evidence.ewf.ewf.Segment")
def test_ewf_open_segment(MockSegment: MagicMock, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(ewf, "MAX_OPEN_SEGMENTS", 2)

    mock_segment = MockSegment.return_value
    mock_segment.volume.sector_size = 512
    mock_segment.sector_count = 2

    mock_fh = [MagicMock(), MagicMock(), MagicMock(), MagicMock()]
    e = ewf.EWF(mock_fh)

    assert e._segment_offsets == [2, 4, 6]
    assert e._segment_lru == [2, 3]

    tmp = e.segment(0)
    assert tmp.offset == 0
    assert tmp.sector_offset == 0
    assert e._segment_lru == [3, 0]

    tmp = e.segment(1)
    assert tmp.offset == 1024
    assert tmp.sector_offset == 2
    assert e._segment_lru == [0, 1]



class Error(Exception):
    pass


class FileNotFoundError(Error, FileNotFoundError):
    pass


class IsADirectoryError(Error, IsADirectoryError):
    pass


class NotADirectoryError(Error, NotADirectoryError):
    pass


class NotASymlinkError(Error):
    pass


class EWFError(Error):
    """Related to EWF (Expert Witness disk image Format)"""


class InvalidSnapshot(Error):
    """Related to ASDF (Acquire Snapshot Data Format)"""


class InvalidBlock(Error):
    """Related to ASDF (Acquire Snapshot Data Format)"""


class UnsupportedVersion(Error):
    """Related to ASDF (Acquire Snapshot Data Format)"""



import hashlib
import io
import struct
from typing import BinaryIO
from zlib import compressobj, crc32


class SubStreamBase(io.RawIOBase):
    """Convenience class for easy sub stream subclassing.

    Additionally adds the `finalize` method.

    Args:
        fh: The file-like object to wrap.
    """

    def __init__(self, fh: BinaryIO):
        self.fh = fh

    def write(self, b: bytes) -> int:
        return self.fh.write(b)

    def tell(self) -> int:
        return self.fh.tell()

    def seek(self, pos: int, whence: int = io.SEEK_CUR) -> int:
        return self.fh.seek(pos, whence)

    def close(self) -> None:
        super().close()

    def finalize(self) -> None:
        self.fh.flush()
        if hasattr(self.fh, "finalize"):
            self.fh.finalize()


class Crc32Stream(SubStreamBase):
    """Compute a CRC32 over all written data.

    This assumes that all data is written as a continuous stream.

    Args:
        fh: The file-like object to wrap.
    """

    def __init__(self, fh: BinaryIO):
        super().__init__(fh)
        self.crc = 0

    def write(self, b: bytes) -> int:
        self.crc = crc32(b, self.crc) & 0xFFFFFFFF
        return self.fh.write(b)

    def digest(self) -> bytes:
        return struct.pack("<I", self.crc)

    def finalize(self) -> None:
        self.fh.write(self.digest())
        super().finalize()


class HashedStream(SubStreamBase):
    """Compute a hash over all written data.

    This assumes that all data is written as a continuous stream.

    Args:
        fh: The file-like object to wrap.
        alg: The hashing algorithm to use. Must be supported by hashlib.
    """

    def __init__(self, fh: BinaryIO, alg: str = "sha256"):
        super().__init__(fh)
        self.ctx = hashlib.new(alg)

    def write(self, b: bytes) -> int:
        self.ctx.update(b)
        return self.fh.write(b)

    def digest(self) -> bytes:
        return self.ctx.digest()

    def hexdigest(self) -> str:
        return self.ctx.hexdigest()

    def close(self) -> None:
        super().close()
        self.fh.close()


class CompressedStream(SubStreamBase):
    """Compress data as it's being written.

    This assumes that all data is written as a continuous stream.

    Args:
        fh: The file-like object to wrap.
    """

    def __init__(self, fh: BinaryIO):
        super().__init__(fh)
        self.cobj = compressobj()

    def write(self, b: bytes) -> int:
        return self.fh.write(self.cobj.compress(b))

    def finalize(self) -> None:
        self.fh.write(self.cobj.flush())
        super().finalize()



from __future__ import annotations

from dissect.cstruct import cstruct

asdf_def = """
flag FILE_FLAG : uint32 {
    SHA256      = 0x01,
};

flag BLOCK_FLAG : uint8 {
    CRC32       = 0x01,
    COMPRESS    = 0x02,
};

struct header {
    char        magic[4];       // File magic, must be "ASDF"
    FILE_FLAG   flags;          // File flags
    uint8       version;        // File version
    char        reserved1[7];   // Reserved
    uint64      timestamp;      // Creation timestamp of the file
    char        reserved2[8];   // Reserved
    char        guid[16];       // GUID, should be unique per writer
};

struct block {
    char        magic[4];       // Block magic, must be "BL\\xa5\\xdf"
    BLOCK_FLAG  flags;          // Block flags
    uint8       idx;            // Stream index, some reserved values have special meaning
    char        reserved[2];    // Reserved
    uint64      offset;         // Absolute offset of block in stream
    uint64      size;           // Size of block in stream
};

struct table_entry {
    BLOCK_FLAG  flags;          // Block flags
    uint8       idx;            // Stream index, some reserved values have special meaning
    char        reserved[2];    // Reserved
    uint64      offset;         // Absolute offset of block in stream
    uint64      size;           // Size of block in stream
    uint64      file_offset;    // Absolute offset of block in file
    uint64      file_size;      // Size of block in file
};

struct footer {
    char        magic[4];       // Footer magic, must be "FT\\xa5\\xdf"
    char        reserved[4];    // Reserved
    uint64      table_offset;   // Offset in file to start of block table
    char        sha256[32];     // SHA256 of this file up until this hash
};
"""

c_asdf = cstruct().load(asdf_def)



from __future__ import annotations

from dissect.cstruct import cstruct

adcrypt_def = """
#define ADCRYPT_MAGIC ADCRYPT\00

enum EncAlgo : uint32 {
    AES128 = 0x1,
    AES192 = 0x2,
    AES256 = 0x3,
};

enum HashAlgo : uint32 {
    SHA256 = 0x1,
    SHA512 = 0x2,
};

struct Header {
    char        magic[8];           // b"ADCRYPT\x00"
    uint32      version;            // 0x01
    uint32      header_size;        // 0x200 aka offset enc data
    int16       num_passwords;      // always -0x1 ?
    int16       num_raw_keys;       // always -0x1 ?
    int16       num_certificates;   // always -0x1 ?
    char        reserved[2];        // 00 00
    EncAlgo     enc_algo;
    HashAlgo    hash_algo;
    uint32      pbkdf2_count;
    uint32      salt_len;
    uint32      key_len;
    uint32      hmac_len;
    char        enc_salt[salt_len];
    char        enc_key[key_len];
    char        hmac_enc_key[hmac_len];
    // padding until 0x200
};
"""

c_adcrypt = cstruct().load(adcrypt_def)



from __future__ import annotations

import io
from typing import BinaryIO

from dissect.util.stream import AlignedStream

try:
    from Crypto.Cipher import AES
    from Crypto.Util import Counter

    HAS_CRYPTO = True
except ImportError:
    HAS_CRYPTO = False


class ADCryptStream(AlignedStream):
    def __init__(self, fh: BinaryIO, key: bytes, index: int):
        if not HAS_CRYPTO:
            raise RuntimeError("Missing required dependency 'pycryptodome' for ADCRYPT decryption")

        self.fh = fh
        self.key = key
        self.index = index

        self.fh.seek(0, io.SEEK_END)
        size = self.fh.tell() - (512 if index == 0 else 0)  # Skip ADCRYPT header
        super().__init__(size)

    def _read(self, offset: int, length: int) -> bytes:
        self.fh.seek(offset + (512 if self.index == 0 else 0))  # Skip ADCRYPT header
        buf = self.fh.read(length)

        ctr = Counter.new(
            128,
            initial_value=self.index << 64 | (offset // (128 // 8)),
            little_endian=True,
        )
        cipher = AES.new(self.key, AES.MODE_CTR, counter=ctr)
        return cipher.decrypt(buf)



from __future__ import annotations

from dissect.evidence.ad1.ad1 import AD1, FileEntry, FileMeta, LogicalImage, MetaType, SegmentFile
from dissect.evidence.ad1.stream import AD1Stream, FileStream
from dissect.evidence.exception import (
    Error,
    FileNotFoundError,
    NotADirectoryError,
    NotASymlinkError,
)

__all__ = [
    "AD1",
    "AD1Stream",
    "Error",
    "FileEntry",
    "FileMeta",
    "FileNotFoundError",
    "FileStream",
    "LogicalImage",
    "MetaType",
    "NotADirectoryError",
    "NotASymlinkError",
    "SegmentFile",
]



from __future__ import annotations

import re
from datetime import datetime, timezone
from functools import cached_property
from pathlib import Path, PurePath, PurePosixPath, PureWindowsPath
from typing import TYPE_CHECKING, BinaryIO

from dissect.evidence.ad1.c_ad1 import c_ad1
from dissect.evidence.ad1.stream import AD1Stream, FileStream
from dissect.evidence.adcrypt.adcrypt import ADCrypt, is_adcrypt
from dissect.evidence.exception import FileNotFoundError, NotADirectoryError, NotASymlinkError

if TYPE_CHECKING:
    from collections.abc import Iterator

EntryType = c_ad1.EntryType
MetaType = c_ad1.MetaType
FileClassType = c_ad1.FileClassType

MAX_OPEN_SEGMENTS = 128


def find_files(path: Path) -> list[Path]:
    files = set()
    for file in path.parent.iterdir():
        if file.stem == path.stem and re.match(r"^\.ad[0-9]+$", file.suffix.lower()):
            files.add(file)
    return sorted(files, key=lambda file: int(file.suffix[3:]))


class AD1:
    """AccessData Logical Image (AD1v4) implementation.

    Supports ``zlib`` compressed images and ADCRYPT encrypted images.

    Should be initialized using a list of segment paths or file-like objects, e.g.::

        fs = AD1([Path("file.ad1"), Path("file.ad2")])
        fs = AD1([Path("file.ad1").open("rb"), Path("file.ad2").open("rb")])

    If the AD1 container is ADCRYPT encrypted, it can be unlocked using either a passphrase or private key::

        fs.unlock(passphrase="my secret passphrase")
        fs.unlock(private_key=Path("path/to/private/key.pem"))

    Resources:
        - Reverse engineering FTK Imager
        - https://github.com/pcbje/pyad1/blob/master/documentation/AccessData%20Format%20(AD1).asciidoc
        - https://github.com/al3ks1s/AD1-tools
        - https://web.archive.org/web/20231013073319/https://tmairi.github.io/posts/dissecting-the-ad1-file-format/
        - https://al3ks1s.fr/posts/adventures-part-1/
    """

    def __init__(self, fh: BinaryIO | list[BinaryIO]):
        fhs = [fh] if not isinstance(fh, list) else fh
        self.fh = fhs
        self.root = VirtualEntry(self, "/")

        self._segments: dict[int, SegmentFile] = {}
        self._segment_lru = []
        self._segment_offsets = []

        self.size = 0
        self.stream: AD1Stream | None = None
        self.logical_image: LogicalImage | None = None

        if not self.fh:
            raise ValueError("No segment files provided for AD1 container")

        self.adcrypt = None

        first_segment = self.segment(0)
        if is_adcrypt(first_segment.fh):
            self.adcrypt = ADCrypt(first_segment.fh)
        else:
            self._open_ad1()

    def is_adcrypt(self) -> bool:
        """Return whether the AD1 container is ADCRYPT encrypted."""
        return self.adcrypt is not None

    def is_locked(self) -> bool:
        """Return whether the ADCRYPT container is locked."""
        return self.is_adcrypt() and self.adcrypt.is_locked()

    def segment(self, idx: int) -> SegmentFile:
        """Open a segment by index.

        Implements a simple LRU cache to limit the number of open segments.

        Args:
            idx: Index or URI of the segment to open.
        """
        # Poor mans LRU
        if idx in self._segments:
            self._segment_lru.remove(idx)
            self._segment_lru.append(idx)
            return self._segments[idx]

        if len(self._segment_lru) >= MAX_OPEN_SEGMENTS:
            oldest_idx = self._segment_lru.pop(0)
            oldest_segment = self._segments.pop(oldest_idx)

            # Don't close it if we received it as a file-like object
            if not hasattr(self.fh[oldest_idx], "read"):
                oldest_segment.fh.close()

            del oldest_segment

        fh = self.fh[idx]
        if not hasattr(fh, "read"):
            fh = fh.open("rb") if isinstance(fh, Path) else Path(fh).open("rb")  # noqa: SIM115

        if self.is_adcrypt() and not self.is_locked():
            fh = self.adcrypt.wrap(fh, idx)

        segment = SegmentFile(fh)

        self._segments[idx] = segment
        self._segment_lru.append(idx)

        return segment

    def unlock(self, *, passphrase: str | bytes | None = None, private_key: Path | bytes | None = None) -> None:
        """Unlock the ADCRYPT container with a given passphrase or private key.

        Args:
            passphrase: The passphrase to unlock the container.
            private_key: The private key to unlock the container.

        Raises:
            RuntimeError: If required dependencies are missing.
            ValueError: If unlocking failed.
        """
        self.adcrypt.unlock(passphrase=passphrase, private_key=private_key)

        # Reset LRU
        self._segments = {}
        self._segment_lru = []

        # Open the AD1
        self._open_ad1()

    def _open_ad1(self) -> None:
        self._segment_offsets = []

        offset = 0
        for i in range(len(self.fh)):
            segment = self.segment(i)
            if segment.header.magic != c_ad1.ADSEGMENTEDFILE_MAGIC.encode():
                raise ValueError(f"Invalid AD1 segment file magic in segment {i}")

            if segment.number != i + 1:
                raise ValueError(f"Invalid AD1 segment number in segment {i}, got {segment.number}, expected {i + 1}")

            offset += segment.size
            self._segment_offsets.append(offset)

        self.size = offset
        self.stream = AD1Stream(self)

        # The first .ad1 file contains a logical image header
        first_segment = self.segment(0)
        first_segment.fh.seek(first_segment.header.logical_image_offset)
        self.logical_image = LogicalImage(first_segment.fh)

        # We need to create some fake entries for all parts leading up to `logical_image.name`
        # This name commonly contains the full path each entry in the container is relative to
        _hallicinate_root_entries(self)

    def entry(self, path: str, entry: FileEntry | None = None) -> FileEntry:
        """Return a :class:`FileEntry` based on the given absolute ``path``.

        Args:
            path: Absolute path within the AD1 container.
            entry: The starting entry for relative paths. Defaults to the root entry.

        Raises:
            ValueError: If the ADCRYPT container is locked.
            FileNotFoundError if the given ``path`` is not found in the container.

        Returns:
            :class:`FileEntry` when the given ``path`` is found.
        """
        if self.is_locked():
            raise ValueError("AD1 container is locked by ADCRYPT")

        entry = entry or self.root

        for part in path.split("/"):
            if not part:
                continue

            for child in entry.iterdir():
                if child.name == part and child.type != EntryType.Deleted:
                    entry = child
                    break
            else:
                raise FileNotFoundError(f"File not found: {path}")

        return entry

    def get(self, path: str) -> FileEntry:
        """Shortcut for ``AD1.entry(path)``."""
        return self.entry(path)

    def open(self, path: str) -> FileStream:
        """Shortcut for ``AD1.entry(path).open()``."""
        return self.entry(path).open()


def _hallicinate_root_entries(ad1: AD1) -> None:
    # We need to create some fake entries for all parts leading up to `logical_image.name`
    # This name commonly contains the full path each entry in the container is relative to
    # Not always though, so do some poor mans heuristics
    root_name = ad1.logical_image.name
    if root_name == "Custom Content Image([Multi])":
        ad1.root.entry.child = ad1.logical_image.header.first_file_offset
        if len(ad1.root.children) != 1:
            raise ValueError("Unexpected number of root children for Custom Content Image([Multi])")

        root_name = ad1.root.children[0].name.split(":", 1)[-1]
        first_file_entry = ad1.root.children[0].entry
        is_multi_image = True
    else:
        is_multi_image = False

    root_path = PureWindowsPath(root_name) if "/" not in root_name and "\\" in root_name else PurePosixPath(root_name)
    parts = list(root_path.parts)
    parent = _create_root_entries(ad1, ad1.root, root_path, parts)

    if is_multi_image:
        parent.name = parent.name
        parent.offset = ad1.logical_image.header.first_file_offset
        parent.entry = first_file_entry
    else:
        # Add the first file offset as the first child offset of the last root part
        parent.entry.child = ad1.logical_image.header.first_file_offset


def _create_root_entries(ad1: AD1, parent: FileEntry, path: PurePath, parts: list[str]) -> FileEntry:
    while parts:
        part = parts.pop(0)
        if path.drive and part == f"{path.drive}\\":
            part = path.drive

        entry = VirtualEntry(ad1, part)
        parent.children = [entry]
        parent = entry

    return parent


class SegmentFile:
    """Represents an AD1 segmented file."""

    def __init__(self, fh: BinaryIO):
        self.fh = fh
        self.fh.seek(0)
        self.header = c_ad1.SegmentedFileHeader(self.fh)
        self.number = self.header.segment_number
        self.count = self.header.segment_count
        self.size = self.header.segment_size - 512  # Subtract header size

    def __repr__(self) -> str:
        return f"<SegmentFile number={self.number} count={self.count} size={self.size}>"


class LogicalImage:
    """Represents an AD1 logical image."""

    def __init__(self, fh: BinaryIO):
        self.fh = fh
        self.header = c_ad1.LogicalImageHeader(fh)
        self.name = self.header.name.decode()
        self.version = self.header.version
        self.offset = self.header.first_file_offset
        self.chunk_size = self.header.chunk_size

    def __repr__(self) -> str:
        return (
            f"<LogicalImage version={self.version} name={self.name} offset={self.offset} chunk_size={self.chunk_size}>"
        )


class FileEntry:
    """Represents a file entry in an AD1 logical image."""

    def __init__(self, ad1: AD1, offset: int):
        self.ad1 = ad1
        self.offset = offset

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} type={self.type.name} name={self.name!r} size={self.size}>"

    @cached_property
    def entry(self) -> c_ad1.FileEntry:
        self.ad1.stream.seek(self.offset)
        return c_ad1.FileEntry(self.ad1.stream)

    @cached_property
    def name(self) -> str:
        return self.entry.name.decode()

    @cached_property
    def type(self) -> EntryType:
        return self.entry.type

    @cached_property
    def meta(self) -> dict[MetaType, FileMeta]:
        result = {}

        offset = self.entry.meta
        while offset != 0:
            meta = FileMeta(self.ad1, offset)
            offset = meta.next
            result[meta.type] = meta

        return result

    @cached_property
    def children(self) -> list[FileEntry]:
        result = []

        offset = self.entry.child
        while offset != 0:
            child = FileEntry(self.ad1, offset)
            result.append(child)
            offset = child.entry.next

        return result

    @cached_property
    def size(self) -> int:
        if meta := self.meta.get(MetaType.FileSize):
            return int(meta.data)
        return 0

    @cached_property
    def atime(self) -> datetime:
        if meta := self.meta.get(MetaType.DateAccessed):
            return convert_ts(meta.data)
        return datetime(1970, 1, 1, 0, 0, 0, tzinfo=timezone.utc)

    @cached_property
    def ctime(self) -> datetime:
        if meta := self.meta.get(MetaType.MFTFileDateChanged, self.meta.get(MetaType.DateModified)):
            return convert_ts(meta.data)
        return datetime(1970, 1, 1, 0, 0, 0, tzinfo=timezone.utc)

    @cached_property
    def mtime(self) -> datetime:
        if meta := self.meta.get(MetaType.DateModified):
            return convert_ts(meta.data)
        return datetime(1970, 1, 1, 0, 0, 0, tzinfo=timezone.utc)

    @cached_property
    def btime(self) -> datetime:
        if meta := self.meta.get(MetaType.DateCreated):
            return convert_ts(meta.data)
        return datetime(1970, 1, 1, 0, 0, 0, tzinfo=timezone.utc)

    @cached_property
    def md5(self) -> str | None:
        if meta := self.meta.get(MetaType.MD5):
            return meta.data.decode()
        return None

    @cached_property
    def sha1(self) -> str | None:
        if meta := self.meta.get(MetaType.SHA1):
            return meta.data.decode()
        return None

    def is_file(self) -> bool:
        return self.type in (EntryType.File, EntryType.Unknown_File)

    def is_dir(self) -> bool:
        return self.type == EntryType.Directory

    def is_symlink(self) -> bool:
        if meta := self.meta.get(MetaType.FileClass):
            return int.from_bytes(meta.data, "little") == FileClassType.ReparsePoint
        return False

    def listdir(self) -> list[str]:
        if not self.is_dir():
            raise NotADirectoryError(self.name)
        return [child.name for child in self.children]

    def iterdir(self) -> Iterator[FileEntry]:
        if not self.is_dir():
            raise NotADirectoryError(self.name)
        yield from self.children

    def readlink(self) -> str:
        if not self.is_symlink():
            raise NotASymlinkError(self.name)

        # TODO: Investigate symlinks for unix-like filesystems.

        reparse_point = c_ad1.ReparsePoint(self.open())
        return reparse_point.link.strip("\00").split("\00")[-1]

    def open(self) -> FileStream:
        """Open the file entry for reading."""
        if self.is_dir():
            raise IsADirectoryError(self.name)
        return FileStream(self)


class VirtualEntry(FileEntry):
    """Represents the root entry in an AD1 logical image."""

    def __init__(self, ad1: AD1, name: str):
        super().__init__(ad1, -1)
        self.entry = c_ad1.FileEntry(name=name.encode(), type=EntryType.Directory, size=0)


class FileMeta:
    """Represents a single AD1 logical file metadata item found inside :class:`FileEntry`."""

    def __init__(self, ad1: AD1, offset: int):
        self.ad1 = ad1
        self.offset = offset

        self.ad1.stream.seek(offset)
        self.entry = c_ad1.FileMeta(self.ad1.stream)

        self.category = self.entry.category
        self.type = self.entry.type
        self.data = self.entry.data

        self.next = self.entry.next

    def __repr__(self) -> str:
        return f"<FileMeta category={self.category} type={self.type} data={self.data}>"


def convert_ts(value: bytes) -> datetime:
    """Convert an AD1 timestamp to datetime object. Assuming this is UTC."""

    # DateCreated does not (always) have ``.%f`` precision.
    fmt = "%Y%m%dT%H%M%S.%f" if b"." in value else "%Y%m%dT%H%M%S"
    return datetime.strptime(value.decode(), fmt).replace(tzinfo=timezone.utc)



# Generated by cstruct-stubgen
from typing import BinaryIO, Literal, TypeAlias, overload

import dissect.cstruct as __cs__

class _c_adcrypt(__cs__.cstruct):
    ADCRYPT_MAGIC: Literal["ADCRYPT\x00"] = ...
    class EncAlgo(__cs__.Enum):
        AES128 = ...
        AES192 = ...
        AES256 = ...

    class HashAlgo(__cs__.Enum):
        SHA256 = ...
        SHA512 = ...

    class Header(__cs__.Structure):
        magic: __cs__.CharArray
        version: _c_adcrypt.uint32
        header_size: _c_adcrypt.uint32
        num_passwords: _c_adcrypt.int16
        num_raw_keys: _c_adcrypt.int16
        num_certificates: _c_adcrypt.int16
        reserved: __cs__.CharArray
        enc_algo: _c_adcrypt.EncAlgo
        hash_algo: _c_adcrypt.HashAlgo
        pbkdf2_count: _c_adcrypt.uint32
        salt_len: _c_adcrypt.uint32
        key_len: _c_adcrypt.uint32
        hmac_len: _c_adcrypt.uint32
        enc_salt: __cs__.CharArray
        enc_key: __cs__.CharArray
        hmac_enc_key: __cs__.CharArray
        @overload
        def __init__(
            self,
            magic: __cs__.CharArray | None = ...,
            version: _c_adcrypt.uint32 | None = ...,
            header_size: _c_adcrypt.uint32 | None = ...,
            num_passwords: _c_adcrypt.int16 | None = ...,
            num_raw_keys: _c_adcrypt.int16 | None = ...,
            num_certificates: _c_adcrypt.int16 | None = ...,
            reserved: __cs__.CharArray | None = ...,
            enc_algo: _c_adcrypt.EncAlgo | None = ...,
            hash_algo: _c_adcrypt.HashAlgo | None = ...,
            pbkdf2_count: _c_adcrypt.uint32 | None = ...,
            salt_len: _c_adcrypt.uint32 | None = ...,
            key_len: _c_adcrypt.uint32 | None = ...,
            hmac_len: _c_adcrypt.uint32 | None = ...,
            enc_salt: __cs__.CharArray | None = ...,
            enc_key: __cs__.CharArray | None = ...,
            hmac_enc_key: __cs__.CharArray | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

# Technically `c_adcrypt` is an instance of `_c_adcrypt`, but then we can't use it in type hints
c_adcrypt: TypeAlias = _c_adcrypt



from __future__ import annotations

from dissect.evidence.ad1.ad1 import AD1
from dissect.evidence.asdf.asdf import AsdfSnapshot, AsdfStream
from dissect.evidence.ewf.ewf import EWF

__all__ = [
    "AD1",
    "EWF",
    "AsdfSnapshot",
    "AsdfStream",
]
